from __future__ import annotations

import os
from functools import lru_cache
from typing import Callable, Optional

import firefly as ff

from ...domain.resource_name_aware import ResourceNameAware
from ...domain.service.execution_context import ExecutionContext
from ...domain.service.resource_monitor import ResourceMonitor

if os.environ.get('ADAPTIVE_MEMORY'):
    @ff.register_middleware(index=0, buses=['event', 'command'])
    class AdaptiveMemoryRoutingMiddleware(ff.Middleware, ResourceNameAware, ff.LoggerAware):
        _resource_monitor: ResourceMonitor = None
        _execution_context: ExecutionContext = None
        _message_transport: ff.MessageTransport = None
        _configuration: ff.Configuration = None
        _context: str = None

        def __init__(self):
            context = self._configuration.contexts['firefly_aws']
            if context.get('memory_async') == 'adaptive':
                self._memory_settings = sorted(list(map(int, context.get('memory_settings'))))
                if self._memory_settings is None:
                    raise ff.ConfigurationError(
                        'When using "adaptive" memory you must provide a list of memory_settings'
                    )
                self._memory_mappings = self._configuration.contexts[self._context].get('memory_settings', {})

        def __call__(self, message: ff.Message, next_: Callable) -> Optional[ff.Message]:
            function_name = self._lambda_function_name(self._context, 'Async')
            if self._execution_context.context and self._execution_context.context.function_name == function_name:
                if str(message) in self._memory_mappings:
                    setattr(message, '_memory', self._memory_mappings[str(message)])
                elif 'default' in self._memory_mappings:
                    setattr(message, '_memory', self._memory_mappings['default'])

                if not hasattr(message, '_memory'):
                    memory = self._get_memory_level(str(message))
                    if memory is None:
                        self._resource_monitor.set_memory_level(str(message), self._memory_settings[0])
                        setattr(message, '_memory', str(self._memory_settings[0]))
                        self._get_memory_level.cache_clear()
                    else:
                        setattr(message, '_memory', self._get_memory_level(str(message)))

                self.info('Routing message %s to %s', str(message), self._lambda_function_name(
                    self._context, type_='Async', memory=getattr(message, '_memory')
                ))
                self._enqueue_message(message)

                return

            else:
                return next_(message)

        def _enqueue_message(self, message: ff.Message):
            if isinstance(message, ff.Event):
                self._message_transport.dispatch(message)
            elif isinstance(message, ff.Command):
                setattr(message, '_async', True)
                self._message_transport.invoke(message)

        @lru_cache(maxsize=None)
        def _get_memory_level(self, message: str):
            return self._resource_monitor.get_memory_level(message)



from __future__ import annotations

import firefly as ff


class PrepareS3Download(ff.DomainService):
    _s3_client = None

    def __call__(self, bucket: str, key: str):
        return self._s3_client.generate_presigned_url('get_object', Params={'Bucket': bucket, 'Key': key})



#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import gzip
import uuid

import firefly as ff
import firefly_aws.domain as awsd
from botocore.exceptions import ClientError


class BotoS3Service(awsd.S3Service, ff.LoggerAware):
    _configuration: ff.Configuration = None
    _s3_client = None
    _bucket: str = None

    def store_download(self, data: str, extension: str = None, file_name: str = None, apply_compression: bool = True):
        key = file_name if file_name is not None else str(uuid.uuid4())
        content_encoding = None
        if extension is not None:
            key += f".{extension}"

        if apply_compression:
            data = gzip.compress(data.encode('utf-8'))
            key += '.gz'
            content_encoding = 'gzip'

        key = f'/tmp/{key}'

        params = {
            'Body': data,
            'Bucket': self._bucket,
            'Key': key,
        }
        if content_encoding is not None:
            params['Metadata'] = {
                'ContentEncoding': content_encoding,
            }

        self._s3_client.put_object(**params)

        return self._s3_client.generate_presigned_url(
            'get_object', Params={'Bucket': self._bucket, 'Key': key}
        )



from __future__ import annotations

import firefly as ff


@ff.command_handler()
class UpdateResourceSettings(ff.ApplicationService):
    def __call__(self, **kwargs):
        print("Updating resource settings")
        print(kwargs)



from __future__ import annotations

from datetime import datetime, timedelta
from time import sleep

import firefly as ff
from botocore.exceptions import ClientError


class DdbRateLimiter(ff.RateLimiter):
    _ddb_client = None
    _ddb_table: str = None

    def acquire(self, key: str, max_concurrent: int, timeout: int = None) -> bool:
        mark = datetime.now()
        while True:
            try:
                self._ddb_client.update_item(
                    TableName=self._ddb_table,
                    Key={
                        'pk': {'S': key},
                        'sk': {'S': 'rate-limiter'}
                    },
                    UpdateExpression='ADD Rate :inc SET TimeToLive = :ttl',
                    ConditionExpression='attribute_not_exists(Rate) OR Rate < :limit',
                    ExpressionAttributeValues={
                        ':limit': {'N': str(max_concurrent)},
                        ':inc': {'N': '1'},
                        ':ttl': {'N': str(int((datetime.now() + timedelta(minutes=5)).timestamp()))}
                    }
                )
                return True
            except ClientError as e:
                if 'ConditionalCheckFailedException' in str(e):
                    print(e)
                    if timeout is not None and (datetime.now() - mark).seconds >= timeout:
                        raise TimeoutError(f'Could not acquire mutex for key: {key}')
                    sleep(1)
                else:
                    raise e

    def release(self, key: str):
        self._ddb_client.update_item(
            TableName=self._ddb_table,
            Key={
                'pk': {'S': key},
                'sk': {'S': 'rate-limiter'}
            },
            UpdateExpression='ADD Rate :dec',
            ConditionExpression='Rate > :limit',
            ExpressionAttributeValues={
                ':limit': {'N': '0'},
                ':dec': {'N': '-1'},
            }
        )



from __future__ import annotations

import firefly as ff


class DataApi(ff.LoggerAware):
    _rds_data_client = None
    _db_arn: str = None
    _db_secret_arn: str = None
    _db_name: str = None

    def __init__(self, db_arn: str = None, db_secret_arn: str = None, db_name: str = None):
        if db_arn is not None and db_secret_arn is not None and db_name is not None:
            self._db_arn = db_arn
            self._db_secret_arn = db_secret_arn
            self._db_name = db_name

    def execute(self, sql: str, params: list = None, db_arn: str = None, db_secret_arn: str = None,
                db_name: str = None):
        params = params or []
        self.info('%s - %s', sql, str(params))

        return self._rds_data_client.execute_statement(
            resourceArn=(db_arn or self._db_arn),
            secretArn=(db_secret_arn or self._db_secret_arn),
            database=(db_name or self._db_name),
            includeResultMetadata=True,
            sql=sql,
            parameters=params
        )



#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import shutil
from datetime import datetime
from pprint import pprint
from time import sleep

import firefly as ff
import firefly.infrastructure as ffi
import inflection
import yaml
from botocore.exceptions import ClientError
from troposphere import Template, GetAtt, Ref, Parameter, Output, Export, ImportValue, Join
from troposphere.apigatewayv2 import Api, Stage, Deployment, Integration, Route
from troposphere.awslambda import Function, Code, VPCConfig, Environment, Permission, EventSourceMapping
from troposphere.constants import NUMBER
from troposphere.dynamodb import Table, AttributeDefinition, KeySchema, TimeToLiveSpecification
from troposphere.events import Target, Rule
from troposphere.iam import Role, Policy
from troposphere.s3 import Bucket, LifecycleRule, LifecycleConfiguration
from troposphere.sns import Topic, SubscriptionResource
from troposphere.sqs import Queue, QueuePolicy, RedrivePolicy
import troposphere.kinesis as kinesis
import troposphere.kinesisanalyticsv2 as analytics

from firefly_aws import S3Service, ResourceNameAware


@ff.agent('aws')
class AwsAgent(ff.Agent, ResourceNameAware, ff.LoggerAware):
    _configuration: ff.Configuration = None
    _context_map: ff.ContextMap = None
    _registry: ff.Registry = None
    _s3_client = None
    _s3_service: S3Service = None
    _sns_client = None
    _cloudformation_client = None
    _adaptive_memory = None

    def __init__(self, account_id: str):
        self._account_id = account_id

    def __call__(self, deployment: ff.Deployment, **kwargs):
        self._env = deployment.environment
        try:
            self._bucket = self._configuration.contexts.get('firefly_aws').get('bucket')
        except AttributeError:
            raise ff.FrameworkError('No deployment bucket configured in firefly_aws')

        memory_settings = self._configuration.contexts.get('firefly_aws').get('memory_settings')
        if memory_settings is not None:
            self._adaptive_memory = '1'
            os.environ['ADAPTIVE_MEMORY'] = '1'
        else:
            self._adaptive_memory = None

        self._deployment = deployment
        self._project = self._configuration.all.get('project')
        aws_config = self._configuration.contexts.get('firefly_aws')
        self._aws_config = aws_config
        self._region = aws_config.get('region')
        self._security_group_ids = aws_config.get('vpc', {}).get('security_group_ids')
        self._subnet_ids = aws_config.get('vpc', {}).get('subnet_ids')

        self._template_key = f'cloudformation/templates/{inflection.dasherize(self._service_name())}.json'
        self._create_project_stack()

        for service in deployment.services:
            lambda_path = inflection.dasherize(self._lambda_resource_name(service.name))
            template_path = inflection.dasherize(self._service_name(self._context_map.get_context(service.name).name))
            self._code_path = f'lambda/code/{lambda_path}'
            self._code_key = f'{self._code_path}/{datetime.now().isoformat()}.zip'
            self._template_key = f'cloudformation/templates/{template_path}.json'
            self._deploy_service(service)

    def _deploy_service(self, service: ff.Service):
        context = self._context_map.get_context(service.name)
        if self._aws_config.get('image_uri') is None:
            self._package_and_deploy_code(context)

        template = Template()
        template.set_version('2010-09-09')

        memory_size = template.add_parameter(Parameter(
            f'{self._lambda_resource_name(service.name)}MemorySize',
            Type=NUMBER,
            Default=self._aws_config.get('memory_sync', '3008')
        ))

        timeout_gateway = template.add_parameter(Parameter(
            f'{self._lambda_resource_name(service.name)}GatewayTimeout',
            Type=NUMBER,
            Default='30'
        ))

        timeout_async = template.add_parameter(Parameter(
            f'{self._lambda_resource_name(service.name)}AsyncTimeout',
            Type=NUMBER,
            Default='900'
        ))

        role_title = f'{self._lambda_resource_name(service.name)}ExecutionRole'
        role = self._add_role(role_title, template)

        params = {
            'FunctionName': f'{self._service_name(service.name)}Sync',
            'Role': GetAtt(role_title, 'Arn'),
            'MemorySize': Ref(memory_size),
            'Timeout': Ref(timeout_gateway),
            'Environment': self._lambda_environment(context)
        }

        image_uri = self._aws_config.get('image_uri')
        if image_uri is not None:
            params.update({
                'Code': Code(
                    ImageUri=image_uri
                ),
                'PackageType': 'Image',
            })
        else:
            params.update({
                'Code': Code(
                    S3Bucket=self._bucket,
                    S3Key=self._code_key
                ),
                'Runtime': 'python3.7',
                'Handler': 'handlers.main',
            })

        if self._security_group_ids and self._subnet_ids:
            params['VpcConfig'] = VPCConfig(
                SecurityGroupIds=self._security_group_ids,
                SubnetIds=self._subnet_ids
            )
        api_lambda = template.add_resource(Function(
            f'{self._lambda_resource_name(service.name)}Sync',
            **params
        ))

        route = inflection.dasherize(context.name)
        proxy_route = f'{route}/{{proxy+}}'
        template.add_resource(Permission(
            f'{self._lambda_resource_name(service.name)}SyncPermission',
            Action='lambda:InvokeFunction',
            FunctionName=f'{self._service_name(service.name)}Sync',
            Principal='apigateway.amazonaws.com',
            SourceArn=Join('', [
                'arn:aws:execute-api:',
                self._region,
                ':',
                self._account_id,
                ':',
                ImportValue(self._rest_api_reference()),
                '/*/*/',
                route,
                '*'
            ]),
            DependsOn=api_lambda
        ))

        if self._adaptive_memory:
            value = '3008' if not self._adaptive_memory else '256'
            try:
                value = int(self._aws_config.get('memory_async'))
            except ValueError:
                pass
            memory_size = template.add_parameter(Parameter(
                f'{self._lambda_resource_name(service.name)}MemorySizeAsync',
                Type=NUMBER,
                Default=value
            ))

        params = {
            'FunctionName': self._lambda_function_name(service.name, 'Async'),
            'Role': GetAtt(role_title, 'Arn'),
            'MemorySize': Ref(memory_size),
            'Timeout': Ref(timeout_async),
            'Environment': self._lambda_environment(context)
        }

        if image_uri is not None:
            params.update({
                'Code': Code(
                    ImageUri=image_uri
                ),
                'PackageType': 'Image',
            })
        else:
            params.update({
                'Code': Code(
                    S3Bucket=self._bucket,
                    S3Key=self._code_key
                ),
                'Runtime': 'python3.7',
                'Handler': 'handlers.main',
            })

        if self._security_group_ids and self._subnet_ids:
            params['VpcConfig'] = VPCConfig(
                SecurityGroupIds=self._security_group_ids,
                SubnetIds=self._subnet_ids
            )
        async_lambda = template.add_resource(Function(
            self._lambda_resource_name(service.name, type_='Async'),
            **params
        ))

        if self._adaptive_memory:
            self._add_adaptive_memory_functions(template, context, timeout_async, role_title, async_lambda)
            # self._add_adaptive_memory_streams(template, context, async_lambda, role)

        # Timers
        for cls, _ in context.command_handlers.items():
            if cls.has_timer():
                timer = cls.get_timer()
                if timer.environment is not None and timer.environment != self._env:
                    continue
                if isinstance(timer.command, str):
                    timer_name = timer.command
                else:
                    timer_name = timer.command.__name__

                target = Target(
                    f'{self._service_name(service.name)}AsyncTarget',
                    Arn=GetAtt(self._lambda_resource_name(service.name, type_='Async'), 'Arn'),
                    Id=self._lambda_resource_name(service.name, type_='Async'),
                    Input=f'{{"_context": "{context.name}", "_type": "command", "_name": "{cls.__name__}"}}'
                )
                rule = template.add_resource(Rule(
                    f'{timer_name}TimerRule',
                    ScheduleExpression=f'cron({timer.cron})',
                    State='ENABLED',
                    Targets=[target]
                ))
                template.add_resource(Permission(
                    f'{timer_name}TimerPermission',
                    Action='lambda:invokeFunction',
                    Principal='events.amazonaws.com',
                    FunctionName=Ref(async_lambda),
                    SourceArn=GetAtt(rule, 'Arn')
                ))

        integration = template.add_resource(Integration(
            self._integration_name(context.name),
            ApiId=ImportValue(self._rest_api_reference()),
            PayloadFormatVersion='2.0',
            IntegrationType='AWS_PROXY',
            IntegrationUri=Join('', [
                'arn:aws:lambda:',
                self._region,
                ':',
                self._account_id,
                ':function:',
                Ref(api_lambda),
            ]),
        ))

        template.add_resource(Route(
            f'{self._route_name(context.name)}Base',
            ApiId=ImportValue(self._rest_api_reference()),
            RouteKey=f'ANY /{route}',
            AuthorizationType='NONE',
            Target=Join('/', ['integrations', Ref(integration)]),
            DependsOn=integration
        ))

        template.add_resource(Route(
            f'{self._route_name(context.name)}Proxy',
            ApiId=ImportValue(self._rest_api_reference()),
            RouteKey=f'ANY /{proxy_route}',
            AuthorizationType='NONE',
            Target=Join('/', ['integrations', Ref(integration)]),
            DependsOn=integration
        ))

        # Error alarms / subscriptions

        if 'errors' in self._aws_config:
            alerts_topic = template.add_resource(Topic(
                self._alert_topic_name(service.name),
                TopicName=self._alert_topic_name(service.name)
            ))

            if 'email' in self._aws_config.get('errors'):
                for address in self._aws_config.get('errors').get('email').get('recipients').split(','):
                    template.add_resource(SubscriptionResource(
                        self._alarm_subscription_name(context.name),
                        Protocol='email',
                        Endpoint=address,
                        TopicArn=self._alert_topic_arn(context.name),
                        DependsOn=[alerts_topic]
                    ))

        # Queues / Topics

        subscriptions = {}
        for subscription in self._get_subscriptions(context):
            if subscription['context'] not in subscriptions:
                subscriptions[subscription['context']] = []
            subscriptions[subscription['context']].append(subscription)

        dlq = template.add_resource(Queue(
            f'{self._queue_name(context.name)}Dlq',
            QueueName=f'{self._queue_name(context.name)}Dlq',
            VisibilityTimeout=905,
            ReceiveMessageWaitTimeSeconds=20,
            MessageRetentionPeriod=1209600
        ))
        self._queue_policy(template, dlq, f'{self._queue_name(context.name)}Dlq', subscriptions)

        queue = template.add_resource(Queue(
            self._queue_name(context.name),
            QueueName=self._queue_name(context.name),
            VisibilityTimeout=905,
            ReceiveMessageWaitTimeSeconds=20,
            MessageRetentionPeriod=1209600,
            RedrivePolicy=RedrivePolicy(
                deadLetterTargetArn=GetAtt(dlq, 'Arn'),
                maxReceiveCount=1000
            ),
            DependsOn=dlq
        ))
        self._queue_policy(template, queue, self._queue_name(context.name), subscriptions)

        template.add_resource(EventSourceMapping(
            f'{self._lambda_resource_name(context.name)}AsyncMapping',
            BatchSize=1,
            Enabled=True,
            EventSourceArn=GetAtt(queue, 'Arn'),
            FunctionName=self._lambda_function_name(service.name, 'Async'),
            DependsOn=[queue, async_lambda]
        ))
        topic = template.add_resource(Topic(
            self._topic_name(context.name),
            TopicName=self._topic_name(context.name)
        ))

        for context_name, list_ in subscriptions.items():
            if context_name == context.name and len(list_) > 0:
                template.add_resource(SubscriptionResource(
                    self._subscription_name(context_name),
                    Protocol='sqs',
                    Endpoint=GetAtt(queue, 'Arn'),
                    TopicArn=self._topic_arn(context.name),
                    FilterPolicy={
                        '_name': [x['name'] for x in list_],
                    },
                    RedrivePolicy={
                        'deadLetterTargetArn': GetAtt(dlq, 'Arn'),
                    },
                    DependsOn=[queue, dlq, topic]
                ))
            elif len(list_) > 0:
                if context_name not in self._context_map.contexts:
                    self._find_or_create_topic(context_name)
                template.add_resource(SubscriptionResource(
                    self._subscription_name(context.name, context_name),
                    Protocol='sqs',
                    Endpoint=GetAtt(queue, 'Arn'),
                    TopicArn=self._topic_arn(context_name),
                    FilterPolicy={
                        '_name': [x['name'] for x in list_]
                    },
                    RedrivePolicy={
                        'deadLetterTargetArn': GetAtt(dlq, 'Arn'),
                    },
                    DependsOn=[queue, dlq]
                ))

        # DynamoDB Table

        ddb_table = template.add_resource(Table(
            self._ddb_resource_name(context.name),
            TableName=self._ddb_table_name(context.name),
            AttributeDefinitions=[
                AttributeDefinition(AttributeName='pk', AttributeType='S'),
                AttributeDefinition(AttributeName='sk', AttributeType='S'),
            ],
            BillingMode='PAY_PER_REQUEST',
            KeySchema=[
                KeySchema(AttributeName='pk', KeyType='HASH'),
                KeySchema(AttributeName='sk', KeyType='RANGE'),
            ],
            TimeToLiveSpecification=TimeToLiveSpecification(
                AttributeName='TimeToLive',
                Enabled=True
            )
        ))

        template.add_output(Output(
            "DDBTable",
            Value=Ref(ddb_table),
            Description="Document table"
        ))

        for cb in self._pre_deployment_hooks:
            cb(template=template, context=context, env=self._env)

        self.info('Deploying stack')
        self._s3_client.put_object(
            Body=template.to_json(),
            Bucket=self._bucket,
            Key=self._template_key
        )
        url = self._s3_client.generate_presigned_url(
            ClientMethod='get_object',
            Params={
                'Bucket': self._bucket,
                'Key': self._template_key
            }
        )

        stack_name = self._stack_name(context.name)
        try:
            self._cloudformation_client.describe_stacks(StackName=stack_name)
            self._update_stack(self._stack_name(context.name), url)
        except ClientError as e:
            if f'Stack with id {stack_name} does not exist' in str(e):
                self._create_stack(self._stack_name(context.name), url)
            else:
                raise e

        for cb in self._post_deployment_hooks:
            cb(template=template, context=context, env=self._env)

        self._migrate_schema(context)

        self.info('Done')

    def _add_adaptive_memory_streams(self, template, context: ff.Context, lambda_function, role):
        stream_name = self._stream_resource_name(context.name)
        stream = template.add_resource(kinesis.Stream(
            stream_name,
            Name=stream_name,
            ShardCount=1
        ))

        # sql_text = """
        #                 CREATE OR REPLACE STREAM "DESTINATION_STREAM" (
        #                     "rt" TIMESTAMP,
        #                     "message" CHAR(128),
        #                     "up" BIGINT
        #                 );
                        
        #                 CREATE OR REPLACE PUMP "STREAM_PUMP" AS
        #                     INSERT INTO "DESTINATION_STREAM"
        #                         SELECT STREAM
        #                             FLOOR("SOURCE_SQL_STREAM_001".ROWTIME TO HOUR),
        #                             "message",
        #                             MAX(1)
        #                         FROM "SOURCE_SQL_STREAM_001"
        #                         WHERE "event_type" = 'resource-usage'
        #                         GROUP BY FLOOR("SOURCE_SQL_STREAM_001".ROWTIME TO HOUR), "message"
        #                 ;
        #             """

        # """
        #                             CASE WHEN (AVG(memory_usage) + (STDDEV_SAMP(memory_usage) * 2.58)) > (.9 * max_memory) THEN 1 ELSE 0

        #                             AND (
        #                                 (AVG(memory_usage) + (STDDEV_SAMP(memory_usage) * 2.58)) > (.9 * max_memory)
        #                                 OR (AVG(memory_usage) + (STDDEV_SAMP(memory_usage) * 2.58)) < (.8 * COALESCE(prev_memory_tier, 1000000))
        #                             )
        # """

        sql = """
            CREATE OR REPLACE STREAM "DESTINATION_STREAM" (
                "rt" TIMESTAMP,
                "message" VARCHAR(128),
                "up" BIGINT
            );
    
            CREATE OR REPLACE STREAM "METRICS_STREAM" (
                "rt" TIMESTAMP,
                "message" VARCHAR(128),
                "average" DOUBLE,
                "standard_dev" DOUBLE
            );
    
            CREATE OR REPLACE PUMP "METRICS_PUMP" AS
            INSERT INTO "METRICS_STREAM"
                SELECT STREAM
                    FLOOR(s.ROWTIME TO HOUR),
                    "message",
                    AVG("memory_used"),
                    STDDEV_SAMP("memory_used")
                FROM "PwrLabDevIntegrationStream_001" AS s
                GROUP BY FLOOR(s.ROWTIME TO HOUR), "message";
    
    
            CREATE OR REPLACE PUMP "STREAM_PUMP" AS
            INSERT INTO "DESTINATION_STREAM"
                SELECT STREAM
                    m."rt",
                    m."message",
                    MAX(CASE WHEN (m."average" + (m."standard_dev" * 2.58)) > (.9 * s."max_memory") THEN 1 ELSE 0 END)
                FROM "PwrLabDevIntegrationStream_001" AS s
                JOIN "METRICS_STREAM" AS m
                    ON s."message" = m."message" 
                    AND FLOOR(s.ROWTIME TO HOUR) = m."rt"  
                WHERE 
                        ((m."average" + (m."standard_dev" * 2.58)) > (.9 * s."max_memory"))
                        OR ((m."average" + (m."standard_dev" * 2.58)) < (.8 * s."prev_memory_tier"))
                GROUP BY FLOOR(s.ROWTIME TO HOUR), m."rt", m."message", m."average", m."standard_dev", s."max_memory", s."prev_memory_tier";
                            
        """

        analytics_stream = template.add_resource(analytics.Application(
            self._analytics_application_resource_name(context.name),
            ApplicationName=self._analytics_application_resource_name(context.name),
            ApplicationConfiguration=analytics.ApplicationConfiguration(
                ApplicationCodeConfiguration=analytics.ApplicationCodeConfiguration(
                    CodeContent=analytics.CodeContent(
                        TextContent=sql
                    ),
                    CodeContentType="PLAINTEXT"
                ),
                SqlApplicationConfiguration=analytics.SqlApplicationConfiguration(
                    Inputs=[analytics.Input(
                        InputSchema=analytics.InputSchema(
                            RecordColumns=[
                                analytics.RecordColumn(Mapping='event_type', Name='event_type', SqlType='CHAR(64)'),
                                analytics.RecordColumn(Mapping='message', Name='message', SqlType='CHAR(128)'),
                                analytics.RecordColumn(Mapping='memory_used', Name='memory_used', SqlType='NUMERIC'),
                                analytics.RecordColumn(Mapping='run_time', Name='run_time', SqlType='NUMERIC'),
                                analytics.RecordColumn(Mapping='max_memory', Name='max_memory', SqlType='NUMERIC'),
                                analytics.RecordColumn(Mapping='prev_memory_tier', Name='prev_memory_tier', SqlType='NUMERIC'),
                            ],
                            RecordFormat=analytics.RecordFormat(
                                RecordFormatType="JSON"
                            ),
                        ),
                        KinesisStreamsInput=analytics.KinesisStreamsInput(
                            ResourceARN=GetAtt(stream, 'Arn'),
                        ),
                        NamePrefix=stream_name
                    )]
                )
            ),
            RuntimeEnvironment="SQL-1_0",
            ServiceExecutionRole=GetAtt(role, 'Arn'),
            DependsOn=stream
        ))

        template.add_resource(analytics.ApplicationOutput(
            f'{self._analytics_application_resource_name(context.name)}Output',
            ApplicationName=self._analytics_application_resource_name(context.name),
            Output=analytics.Output(
                DestinationSchema=analytics.DestinationSchema(
                    RecordFormatType="JSON"
                ),
                LambdaOutput=analytics.LambdaOutput(
                    ResourceARN=GetAtt(lambda_function, 'Arn')
                )
            ),
            DependsOn=analytics_stream
        ))

    def _add_adaptive_memory_functions(self, template, context: ff.Context, timeout, role_title, async_lambda):
        memory_settings = self._configuration.contexts.get('firefly_aws').get('memory_settings')
        if memory_settings is None:
            raise ff.ConfigurationError('To use adaptive memory, you must provide a list of memory_settings')

        for memory in memory_settings:
            memory_size = template.add_parameter(Parameter(
                f'{self._lambda_resource_name(context.name)}{memory}MemorySize',
                Type=NUMBER,
                Default=str(memory)
            ))

            params = {
                'FunctionName': self._lambda_function_name(context.name, 'Async', memory=memory),
                'Role': GetAtt(role_title, 'Arn'),
                'MemorySize': Ref(memory_size),
                'Timeout': Ref(timeout),
                'Environment': self._lambda_environment(context),
                'DependsOn': async_lambda,
            }

            image_uri = self._aws_config.get('image_uri')
            if image_uri is not None:
                params.update({
                    'Code': Code(
                        ImageUri=image_uri
                    ),
                    'PackageType': 'Image',
                })
            else:
                params.update({
                    'Code': Code(
                        S3Bucket=self._bucket,
                        S3Key=self._code_key
                    ),
                    'Runtime': 'python3.7',
                    'Handler': 'handlers.main',
                })

            if self._security_group_ids and self._subnet_ids:
                params['VpcConfig'] = VPCConfig(
                    SecurityGroupIds=self._security_group_ids,
                    SubnetIds=self._subnet_ids
                )

            adaptive_memory_lambda = template.add_resource(Function(
                self._lambda_resource_name(context.name, memory=memory),
                **params
            ))

            queue = template.add_resource(Queue(
                self._queue_name(context.name, memory=memory),
                QueueName=self._queue_name(context.name, memory=memory),
                VisibilityTimeout=905,
                ReceiveMessageWaitTimeSeconds=20,
                MessageRetentionPeriod=1209600,
                DependsOn=[adaptive_memory_lambda]
            ))
            # self._queue_policy(template, queue, self._queue_name(context.name), subscriptions)

            template.add_resource(EventSourceMapping(
                f'{self._lambda_resource_name(context.name, memory=memory)}AsyncMapping',
                BatchSize=1,
                Enabled=True,
                EventSourceArn=GetAtt(queue, 'Arn'),
                FunctionName=self._lambda_function_name(context.name, 'Async', memory=memory),
                DependsOn=[queue, adaptive_memory_lambda]
            ))

    def _migrate_schema(self, context: ff.Context):
        for entity in context.entities:
            if issubclass(entity, ff.AggregateRoot) and entity is not ff.AggregateRoot:
                try:
                    repository = self._registry(entity)
                    if isinstance(repository, ffi.RdbRepository):
                        repository.migrate_schema()
                except ff.FrameworkError:
                    self.debug('Could not execute ddl for entity %s', entity)

    def _find_or_create_topic(self, context_name: str):
        arn = f'arn:aws:sns:{self._region}:{self._account_id}:{self._topic_name(context_name)}'
        try:
            self._sns_client.get_topic_attributes(TopicArn=arn)
        except ClientError:
            template = Template()
            template.set_version('2010-09-09')
            template.add_resource(Topic(
                self._topic_name(context_name),
                TopicName=self._topic_name(context_name)
            ))
            self.info(f'Creating stack for context "{context_name}"')
            self._create_stack(self._stack_name(context_name), template)

    @staticmethod
    def _get_subscriptions(context: ff.Context):
        ret = []
        for service, event_types in context.event_listeners.items():
            for event_type in event_types:
                if isinstance(event_type, str):
                    context_name, event_name = event_type.split('.')
                else:
                    context_name = event_type.get_class_context()
                    event_name = event_type.__name__
                ret.append({
                    'name': event_name,
                    'context': context_name,
                })

        return ret

    def _package_and_deploy_code(self, context: ff.Context):
        self.info('Setting up build directory')
        if not os.path.isdir('./build'):
            os.mkdir('./build')
        if os.path.isdir('./build/python-sources'):
            shutil.rmtree('./build/python-sources', ignore_errors=True)
        os.mkdir('./build/python-sources')

        self.info('Installing source files')
        # TODO use setup.py instead?
        import subprocess
        subprocess.call([
            'pip', 'install',
            '-r', self._deployment.requirements_file or 'requirements.txt',
            '-t', './build/python-sources'
        ])

        self.info('Packaging artifact')
        subprocess.call(['cp', 'templates/aws/handlers.py', 'build/python-sources/.'])
        os.chdir('./build/python-sources')
        with open('firefly.yml', 'w') as fp:
            fp.write(yaml.dump(self._configuration.all))

        subprocess.call(['find', '.', '-name', '"*.so"', '-exec', 'strip', '{}', ';'])
        subprocess.call(['find', '.', '-name', '"*.so.*"', '-exec', 'strip', '{}', ';'])
        subprocess.call(['find', '.', '-name', '"*.pyc"', '-exec', 'rm', '-Rf', '{}', ';'])
        subprocess.call(['find', '.', '-name', 'tests', '-type', 'd', '-exec', 'rm', '-R', '{}', ';'])
        for package in ('pandas', 'numpy', 'llvmlite'):
            if os.path.isdir(package):
                subprocess.call(['zip', '-r', package, package])
                subprocess.call(['rm', '-Rf', package])

        file_name = self._code_key.split('/')[-1]
        subprocess.call(['zip', '-r', f'../{file_name}', '.'])
        os.chdir('..')

        self.info('Uploading artifact')
        with open(file_name, 'rb') as fp:
            self._s3_client.put_object(
                Body=fp.read(),
                Bucket=self._bucket,
                Key=self._code_key
            )
        os.chdir('..')

        self._clean_up_old_artifacts(context)

    def _clean_up_old_artifacts(self, context: ff.Context):
        response = self._s3_client.list_objects(
            Bucket=self._bucket,
            Prefix=self._code_path
        )

        files = []
        for row in response['Contents']:
            files.append((row['Key'], row['LastModified']))
        if len(files) < 3:
            return

        files.sort(key=lambda i: i[1], reverse=True)
        for key, _ in files[2:]:
            self._s3_client.delete_object(Bucket=self._bucket, Key=key)

    def _create_project_stack(self):
        update = True
        try:
            self._cloudformation_client.describe_stacks(StackName=self._stack_name())
        except ClientError as e:
            if 'does not exist' not in str(e):
                raise e
            update = False

        self.info('Creating project stack')
        template = Template()
        template.set_version('2010-09-09')

        memory_size = template.add_parameter(Parameter(
            f'{self._stack_name()}MemorySize',
            Type=NUMBER,
            Default=self._aws_config.get('memory_sync', '3008')
        ))

        timeout_gateway = template.add_parameter(Parameter(
            f'{self._stack_name()}GatewayTimeout',
            Type=NUMBER,
            Default='30'
        ))

        template.add_resource(Bucket(
            inflection.camelize(inflection.underscore(self._bucket)),
            BucketName=self._bucket,
            AccessControl='Private',
            LifecycleConfiguration=LifecycleConfiguration(Rules=[
                LifecycleRule(Prefix='tmp', Status='Enabled', ExpirationInDays=1)
            ])
        ))

        api = template.add_resource(Api(
            self._rest_api_name(),
            Name=f'{inflection.humanize(self._project)} {inflection.humanize(self._env)} API',
            ProtocolType='HTTP'
        ))

        role_title = f'{self._rest_api_name()}Role'
        self._add_role(role_title, template)

        default_lambda = template.add_resource(Function(
            f'{self._rest_api_name()}Function',
            FunctionName=self._rest_api_name(),
            Code=Code(
                ZipFile='\n'.join([
                    'def handler(event, context):',
                    '    return event'
                ])
            ),
            Handler='index.handler',
            Role=GetAtt(role_title, 'Arn'),
            Runtime='python3.7',
            MemorySize=Ref(memory_size),
            Timeout=Ref(timeout_gateway)
        ))

        integration = template.add_resource(Integration(
            self._integration_name(),
            ApiId=Ref(api),
            IntegrationType='AWS_PROXY',
            PayloadFormatVersion='2.0',
            IntegrationUri=Join('', [
                'arn:aws:lambda:',
                self._region,
                ':',
                self._account_id,
                ':function:',
                Ref(default_lambda),
            ]),
            DependsOn=f'{self._rest_api_name()}Function'
        ))

        template.add_resource(Route(
            self._route_name(),
            ApiId=Ref(api),
            RouteKey='$default',
            AuthorizationType='NONE',
            Target=Join('/', ['integrations', Ref(integration)]),
            DependsOn=[integration]
        ))

        # Deprecated
        template.add_resource(Stage(
            f'{self._rest_api_name()}Stage',
            StageName='v2',
            ApiId=Ref(api),
            AutoDeploy=True
        ))

        # Deprecated
        template.add_resource(Deployment(
            f'{self._rest_api_name()}Deployment',
            ApiId=Ref(api),
            StageName='v2',
            DependsOn=[
                f'{self._rest_api_name()}Stage',
                self._route_name(),
                self._integration_name(),
                self._rest_api_name(),
            ]
        ))

        template.add_resource(Stage(
            f'{self._rest_api_name()}Stage1',
            StageName='api',
            ApiId=Ref(api),
            AutoDeploy=True
        ))

        template.add_resource(Deployment(
            f'{self._rest_api_name()}Deployment1',
            ApiId=Ref(api),
            StageName='api',
            DependsOn=[
                f'{self._rest_api_name()}Stage1',
                self._route_name(),
                self._integration_name(),
                self._rest_api_name(),
            ]
        ))

        template.add_output([
            Output(
                self._rest_api_reference(),
                Export=Export(self._rest_api_reference()),
                Value=Ref(api)
            ),
        ])

        self._s3_client.put_object(
            Body=template.to_json(),
            Bucket=self._bucket,
            Key=self._template_key
        )
        url = self._s3_client.generate_presigned_url(
            ClientMethod='get_object',
            Params={
                'Bucket': self._bucket,
                'Key': self._template_key
            }
        )

        if update:
            self._update_stack(self._stack_name(), url)
        else:
            self._create_stack(self._stack_name(), url)

    def _add_role(self, role_name: str, template):
        return template.add_resource(Role(
            role_name,
            Path='/',
            Policies=[
                Policy(
                    PolicyName='root',
                    PolicyDocument={
                        'Version': '2012-10-17',
                        'Statement': [
                            {
                                'Action': ['logs:*'],
                                'Resource': 'arn:aws:logs:*:*:*',
                                'Effect': 'Allow',
                            },
                            {
                                'Action': [
                                    'athena:*',
                                    'cloudfront:CreateInvalidation',
                                    'dynamodb:*',
                                    'ec2:*NetworkInterface',
                                    'ec2:DescribeNetworkInterfaces',
                                    'glue:*',
                                    'kinesis:*',
                                    'kinesisanalytics:*',
                                    'lambda:InvokeFunction',
                                    'rds-data:*',
                                    's3:*',
                                    'secretsmanager:GetSecretValue',
                                    'sns:*',
                                    'sqs:*',
                                ] + self._aws_config.get('permissions', []),
                                'Resource': '*',
                                'Effect': 'Allow',
                            }
                        ]
                    }
                )
            ],
            AssumeRolePolicyDocument={
                'Version': '2012-10-17',
                'Statement': [{
                    'Action': ['sts:AssumeRole'],
                    'Effect': 'Allow',
                    'Principal': {
                        'Service': ['lambda.amazonaws.com', 'kinesisanalytics.amazonaws.com']
                    }
                }]
            }
        ))

    def _create_stack(self, stack_name: str, template: str):
        self._cloudformation_client.create_stack(
            StackName=stack_name,
            TemplateURL=template,
            Capabilities=['CAPABILITY_IAM']
        )
        self._wait_for_stack(stack_name)

    def _update_stack(self, stack_name: str, template: str):

        try:
            self._cloudformation_client.update_stack(
                StackName=stack_name,
                TemplateURL=template,
                Capabilities=['CAPABILITY_IAM']
            )
            self._wait_for_stack(stack_name)
        except ClientError as e:
            if 'No updates are to be performed' in str(e):
                return
            raise e

    def _wait_for_stack(self, stack_name: str):
        status = self._cloudformation_client.describe_stacks(StackName=stack_name)['Stacks'][0]
        while status['StackStatus'].endswith('_IN_PROGRESS'):
            self.info('Waiting...')
            sleep(5)
            status = self._cloudformation_client.describe_stacks(StackName=stack_name)['Stacks'][0]

    def _lambda_environment(self, context: ff.Context):
        env = ((context.config.get('extensions') or {}).get('firefly_aws') or {}).get('environment')

        defaults = {
            'PROJECT': self._project,
            'FF_ENVIRONMENT': self._env,
            'ACCOUNT_ID': self._account_id,
            'CONTEXT': context.name,
            'REGION': self._region,
            'BUCKET': self._bucket,
            'DDB_TABLE': self._ddb_table_name(context.name),
        }

        if self._adaptive_memory is not None:
            defaults['ADAPTIVE_MEMORY'] = self._adaptive_memory

        if env is not None:
            defaults.update(env)

        if 'SLACK_ERROR_URL' in os.environ:
            defaults['SLACK_ERROR_URL'] = os.environ.get('SLACK_ERROR_URL')

        return Environment(
            'LambdaEnvironment',
            Variables=defaults
        )

    def _queue_policy(self, template: Template, queue, queue_name: str, subscriptions: dict):
        template.add_resource(QueuePolicy(
            f'{queue_name}Policy',
            Queues=[Ref(queue)],
            PolicyDocument={
                'Version': '2008-10-17',
                'Id': f'{queue_name}Policy',
                'Statement': [{
                    'Action': [
                        'sqs:SendMessage',
                    ],
                    'Effect': 'Allow',
                    'Resource': GetAtt(queue, 'Arn'),
                    'Principal': {
                        'AWS': '*',
                    },
                    'Condition': {
                        'ForAnyValue:ArnEquals': {
                            'aws:SourceArn': [
                                self._topic_arn(name) for name in subscriptions.keys()
                            ]
                        }
                    }
                }]
            },
            DependsOn=queue
        ))



# #  Copyright (c) 2020 JD Williams
# #
# #  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
# #  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
# #  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
# #
# #  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# #  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# #  Public License for more details. You should have received a copy of the GNU Lesser General Public
# #  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
# #
# #  You should have received a copy of the GNU General Public License along with Firefly. If not, see
# #  <http://www.gnu.org/licenses/>.
#
# from __future__ import annotations
#
# import base64
# import os
# import pickle
# from typing import List
#
# import firefly as ff
# import inflection
# from troposphere import AWSObject, BaseAWSObject, GetAtt, Ref, Template
# from troposphere.awslambda import Function, Code
# from troposphere.iam import Role, Policy
#
#

class Service:
    pass

# class Service(ff.Entity):
#     id: str = ff.id_(is_uuid=False)
#     template: Template = ff.optional()
#
#     def __post_init__(self):
#         if self.template is None:
#             self.template = Template()
#         else:
#             self.template = pickle.loads(base64.b64decode(self.template))
#
#     def to_dict(self):
#         return {
#             'id': self.id,
#             'template': base64.b16encode(pickle.dumps(self.template)),
#         }
#
#     def get_lambda(self):
#         return self.get_resource(self._lambda_resource_name())
#
#     def create_lambda(self, config: ff.Configuration):
#         role_title = f'{self._lambda_resource_name()}ExecutionRole'
#         self.add_resource(Role(
#             role_title,
#             Path='/',
#             Policies=[
#                 Policy(
#                     PolicyName='root',
#                     PolicyDocument={
#                         'Version': '2012-10-17',
#                         'Statement': [{
#                             'Action': ['logs:*'],
#                             'Resource': 'arn:aws:logs:*:*:*',
#                             'Effect': 'Allow',
#                         }]
#                     }
#                 )
#             ],
#             AssumeRolePolicyDocument={
#                 'Version': '2012-10-17',
#                 'Statement': [{
#                     'Action': ['sts:AssumeRole'],
#                     'Effect': 'Allow',
#                     'Principal': {
#                         'Service': ['lambda.amazonaws.com']
#                     }
#                 }]
#             }
#         ))
#
#         self.add_resource(Function(
#             self._lambda_resource_name(),
#             Code=Code(
#                 S3Bucket='',
#                 S3Key=f'{os.environ.get("ENV")}/lambda/code/{self._lambda_resource_name()}.zip'
#             ),
#             Handler='handlers.main',
#             Role=GetAtt(role_title, 'Arn'),
#             Runtime='Python3.7',
#             MemorySize=Ref
#         ))
#
#     def add_resource(self, config: AWSObject):
#         self.get_next_stack().resources.append(config)
#
#     def add_parameter(self, param: BaseAWSObject):
#         self.get_next_stack().parameters.append(param)
#
#     def get_next_stack(self):
#         i = 0
#         stack = self.stacks[i]
#         while stack.num_resources() >= 200:
#             i += 1
#             stack = self.stacks[i]
#         return stack
#
#     def get_resource(self, title: str):
#         for stack in self.stacks:
#             if stack.has_resource(title):
#                 return stack.get_resource(title)
#
#     def _lambda_resource_name(self):
#         return f'{inflection.camelize(self.id)}Function'



from __future__ import annotations

import firefly as ff
from boto3.dynamodb.types import TypeSerializer, TypeDeserializer


class DdbDeserializer(ff.DomainService):
    _serializer: TypeSerializer = None
    _deserializer: TypeDeserializer = None

    def __init__(self):
        self._serializer = TypeSerializer()
        self._deserializer = TypeDeserializer()

    def serialize(self, data):
        return self._serializer.serialize(data)

    def deserialize(self, data: dict):
        return {
            k: self._deserializer.deserialize(v)
            for k, v in data.items()
        }



#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import uuid
from typing import Any, Union

import firefly as ff
import firefly_aws as domain
from botocore.exceptions import ClientError
from firefly import Query, Command, Event


class BotoMessageTransport(ff.MessageTransport, domain.ResourceNameAware):
    _serializer: ff.Serializer = None
    _store_large_payloads_in_s3: domain.StoreLargePayloadsInS3 = None
    _load_payload: domain.LoadPayload = None
    _lambda_client = None
    _sns_client = None
    _sqs_resource = None
    _s3_client = None
    _bucket: str = None
    _context: str = None

    def dispatch(self, event: Event) -> None:
        try:
            if hasattr(event, '_memory'):
                self._enqueue_message(event, context=self._context)
            else:
                self._sns_client.publish(
                    TopicArn=self._topic_arn(event.get_context()),
                    Message=self._store_large_payloads_in_s3(
                        self._serializer.serialize(event),
                        name=event.__class__.__name__,
                        type_='command' if isinstance(event, ff.Command) else 'event',
                        context=event.get_context(),
                        id_=getattr(event, '_id', str(uuid.uuid4()))
                    ),
                    MessageAttributes={
                        '_name': {
                            'DataType': 'String',
                            'StringValue': event.__class__.__name__,
                        },
                        '_type': {
                            'DataType': 'String',
                            'StringValue': 'event'
                        },
                        '_context': {
                            'DataType': 'String',
                            'StringValue': event.get_context()
                        },
                    }
                )
        except ClientError as e:
            raise ff.MessageBusError(str(e))

    def invoke(self, command: Command) -> Any:
        return self._invoke_lambda(command)

    def request(self, query: Query) -> Any:
        return self._invoke_lambda(query)

    def _invoke_lambda(self, message: Union[Command, Query]):
        if hasattr(message, '_async') and getattr(message, '_async') is True:
            return self._enqueue_message(message)

        try:
            response = ff.retry(
                lambda: self._lambda_client.invoke(
                    FunctionName=f'{self._service_name(message.get_context())}Sync',
                    InvocationType='RequestResponse',
                    LogType='None',
                    Payload=self._serializer.serialize(message)
                ),
                wait=2
            )
        except ClientError as e:
            raise ff.MessageBusError(str(e))

        ret = self._serializer.deserialize(response['Payload'].read().decode('utf-8'))
        if isinstance(ret, dict) and 'PAYLOAD_KEY' in ret:
            ret = self._load_payload(ret['PAYLOAD_KEY'])

        return ret

    def _enqueue_message(self, message: ff.Message, context: str = None):
        memory = None
        if hasattr(message, '_memory'):
            memory = getattr(message, '_memory')
        queue = self._sqs_resource.get_queue_by_name(
            QueueName=self._queue_name(context or message.get_context(), memory=memory)
        )
        queue.send_message(MessageBody=self._store_large_payloads_in_s3(
            self._serializer.serialize(message),
            name=message.__class__.__name__,
            type_='command' if isinstance(message, ff.Command) else 'event',
            context=message.get_context(),
            id_=getattr(message, '_id', str(uuid.uuid4()))
        ))



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any

import firefly as ff
from dynamodb_json import json_util


class DdbCache(ff.Cache):
    _ddb_client = None
    _ddb_table: str = None

    def set(self, key: str, value: Any, ttl: int = None, **kwargs) -> Any:
        if '.' in key:
            return self._set_property(key, value)
        else:
            return self._set_item(key, value, ttl)

    def get(self, key: str, **kwargs) -> Any:
        response = self._ddb_client.get_item(
            TableName=self._ddb_table,
            Key=json_util.dumps({
                'pk': key,
                'sk': 'CacheItem',
            }, as_dict=True),
        )

        if 'Item' in response:
            return json_util.loads(response['Item'], as_dict=True)['value']

        return None

    def delete(self, key: str, **kwargs) -> Any:
        response = self._ddb_client.delete_item(
            TableName=self._ddb_table,
            Key=json_util.dumps({
                'pk': key,
                'sk': 'CacheItem',
            }, as_dict=True),
            ReturnValues='ALL_OLD'
        )

        if 'Attributes' in response:
            return json_util.loads(response['Attributes'], as_dict=True)['value']

        return None

    def clear(self):
        raise NotImplemented('DdbCache does not support clear()')

    def increment(self, key: str, amount: int = 1, **kwargs) -> Any:
        k = key.split('.').pop(0)
        attribute_names, path = self._process_path(key)

        params = {
            'TableName': self._ddb_table,
            'Key': json_util.dumps({
                'pk': k,
                'sk': 'CacheItem',
            }, as_dict=True),
            'UpdateExpression': f'SET {path} = {path} + :inc',
            'ExpressionAttributeValues': json_util.dumps({
                ':inc': amount,
            }, as_dict=True),
            'ExpressionAttributeNames': attribute_names,
            'ReturnValues': 'ALL_NEW',
        }

        ret = self._ddb_client.update_item(**params)
        if 'Attributes' in ret:
            return json_util.loads(ret['Attributes'], as_dict=True)['value']

        return None

    def decrement(self, key: str, amount: int = 1, **kwargs) -> Any:
        return self.increment(key, amount=(-amount))

    def add(self, key: str, value: Any, **kwargs) -> Any:
        k = key.split('.').pop(0)
        attribute_names = {f'#val_{i}': v for i, v in enumerate(key.split('.'))}
        attribute_names['#val_0'] = 'value'
        path = '.'.join([f'#val_{i}' for i, _ in enumerate(key.split('.'))])

        params = {
            'TableName': self._ddb_table,
            'Key': json_util.dumps({
                'pk': k,
                'sk': 'CacheItem',
            }, as_dict=True),
            'UpdateExpression': f'SET {path} = list_append(if_not_exists({path}, :empty_list), :val)',
            'ExpressionAttributeValues': json_util.dumps({
                ':val': [value],
                ':empty_list': [],
            }, as_dict=True),
            'ExpressionAttributeNames': attribute_names,
            'ReturnValues': 'ALL_NEW',
        }

        ret = self._ddb_client.update_item(**params)
        if 'Attributes' in ret:
            return json_util.loads(ret['Attributes'], as_dict=True)['value']

        return None

    def remove(self, key: str, value: Any, **kwargs) -> Any:
        parts = key.split('.')
        item = self.get(parts.pop(0))
        if item is None:
            return None

        while len(parts) > 0:
            item = item[parts.pop(0)]
        index = item.index(value)

        if index is None:
            return None

        k = key.split('.').pop(0)
        attribute_names, path = self._process_path(key)

        params = {
            'TableName': self._ddb_table,
            'Key': json_util.dumps({
                'pk': k,
                'sk': 'CacheItem',
            }, as_dict=True),
            'UpdateExpression': f'REMOVE {path}[{index}]',
            'ExpressionAttributeNames': attribute_names,
            'ReturnValues': 'ALL_NEW',
        }

        ret = self._ddb_client.update_item(**params)
        if 'Attributes' in ret:
            return json_util.loads(ret['Attributes'], as_dict=True)['value']

        return None

    def _set_property(self, key: str, value: Any):
        k = key.split('.').pop(0)
        attribute_names = {f'#val_{i}': v for i, v in enumerate(key.split('.'))}
        attribute_names['#val_0'] = 'value'
        path = '.'.join([f'#val_{i}' for i, _ in enumerate(key.split('.'))])

        params = {
            'TableName': self._ddb_table,
            'Key': json_util.dumps({
                'pk': k,
                'sk': 'CacheItem',
            }, as_dict=True),
            'UpdateExpression': f'SET {path} = :val',
            'ExpressionAttributeValues': json_util.dumps({
                ':val': value,
            }, as_dict=True),
            'ExpressionAttributeNames': attribute_names,
            'ReturnValues': 'ALL_NEW',
        }

        ret = self._ddb_client.update_item(**params)
        if 'Attributes' in ret:
            return json_util.loads(ret['Attributes'], as_dict=True)['value']

        return None

    def _set_item(self, key: str, value: Any, ttl: int = None, if_not_exists: bool = False):
        item = {
            'pk': key,
            'sk': 'CacheItem',
            'value': value,
        }

        if ttl is not None:
            item['TimeToLive'] = (datetime.now() + timedelta(seconds=ttl)).timestamp()

        params = {
            'TableName': self._ddb_table,
            'Item': json_util.dumps(item, as_dict=True),
            'ReturnValues': 'ALL_OLD',
        }

        if if_not_exists is True:
            params['KeyConditionExpression'] = 'if_not_exists(pk)'

        ret = self._ddb_client.put_item(**params)
        if 'Attributes' in ret:
            return json_util.loads(ret['Attributes'], as_dict=True)

        return None

    def _process_path(self, key: str):
        attribute_names = {f'#val_{i}': v for i, v in enumerate(key.split('.'))}
        attribute_names['#val_0'] = 'value'
        path = '.'.join([f'#val_{i}' for i, _ in enumerate(key.split('.'))])

        return attribute_names, path



#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from .agent import *
from .boto_message_transport import BotoMessageTransport
from .boto_s3_service import BotoS3Service
from .cognito_jwt_decoder import CognitoJwtDecoder
from .data_api import DataApi
from .ddb_cache import DdbCache
from .ddb_mutex import DdbMutex
from .ddb_rate_limiter import DdbRateLimiter
from .ddb_resource_monitor import *
from .s3_file_system import S3FileSystem



#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Union, Callable, Optional, List

import firefly as ff
from firefly import domain as ffd
from firefly.domain.repository.repository import T


class CognitoRepository(ff.Repository):
    def __init__(self, cognito_idp_client, **kwargs):
        super().__init__()
        self._client = cognito_idp_client

    def append(self, entity: T, **kwargs):
        pass

    def remove(self, entity: T, **kwargs):
        pass

    def find(self, exp: Union[str, Callable], **kwargs) -> Optional[T]:
        pass

    def filter(self, cb: Callable, **kwargs) -> List[T]:
        pass

    def reduce(self, cb: Callable) -> Optional[T]:
        pass

    def __iter__(self):
        pass

    def __next__(self):
        pass

    def __len__(self):
        pass

    def __getitem__(self, item):
        pass

    def commit(self, **kwargs):
        pass

    def execute_ddl(self):
        pass

    def raw(self, cb: Union[Callable, ffd.BinaryOp] = None, limit: int = None):
        pass



#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Callable, Optional, Union

import firefly as ff
import inflection
from botocore.exceptions import ClientError
from firefly import domain as ffd
from firefly.domain.repository.repository import T


class S3Repository(ff.Repository[T]):
    def __init__(self, s3_client, serializer: ff.Serializer, bucket: str, prefix: str = 'object-store/aggregates'):
        self._s3_client = s3_client
        self._serializer = serializer
        self._bucket = bucket
        name = inflection.pluralize(inflection.dasherize(inflection.underscore(self._type().__name__)))
        self._storage_path = f'{prefix}/{name}'.lstrip('/')

    def add(self, entity: T):
        try:
            self._s3_client.put_object(
                Bucket=self._bucket,
                Key=f'{self._storage_path}/{entity.id_value()}.json',
                Body=self._serializer.serialize(entity.to_dict()),
            )
        except ClientError as e:
            raise ff.RepositoryError(str(e))

    def remove(self, entity: T):
        try:
            self._s3_client.delete_object(
                Bucket=self._bucket,
                Key=f'{self._storage_path}/{entity.id_value()}',
            )
        except ClientError as e:
            raise ff.RepositoryError(str(e))

    def find(self, exp: Union[str, Callable]) -> Optional[T]:
        if isinstance(exp, str):
            try:
                response = self._s3_client.get_object(
                    Bucket=self._bucket,
                    Key=f'{self._storage_path}/{exp}.json',
                )
                data = self._serializer.deserialize(response['Body'].read())
                return self._type().from_dict(data)
            except ClientError as e:
                if 'NoSuchKey' in str(e):
                    return None
                raise ff.RepositoryError(str(e))

        # TODO implement search criteria

    def filter(self, cb: Callable) -> List[T]:
        # TODO implement search criteria
        pass

    def reduce(self, cb: Callable) -> Optional[T]:
        # TODO implement search criteria
        pass

    def append(self, entity: T):
        raise NotImplementedError()

    def commit(self):
        raise NotImplementedError()

    def execute_ddl(self):
        raise NotImplementedError()

    def raw(self, cb: Union[Callable, ffd.BinaryOp] = None, limit: int = None):
        raise NotImplementedError()

    def __iter__(self):
        pass

    def __next__(self):
        pass

    def __len__(self):
        pass

    def __getitem__(self, item):
        pass



from __future__ import annotations

from abc import ABC
from typing import Type

import firefly.domain as ffd
import firefly.infrastructure as ffi
from firefly.infrastructure.repository.rdb_repository import Index, Column

from ..data_api_storage_interface import DataApiStorageInterface


class DataApiMysqlBase(DataApiStorageInterface, ffi.LegacyStorageInterface, ABC):
    def _get_table_indexes(self, entity: Type[ffd.Entity]):
        result = self._execute(*self._generate_query(entity, 'mysql/get_indexes.sql'))
        indexes = {}
        if result:
            for row in result:
                name = row['INDEX_NAME']
                if name == 'PRIMARY':
                    continue
                column = row['COLUMN_NAME']
                if name not in indexes:
                    indexes[name] = Index(name=name, columns=[column], unique=row['NON_UNIQUE'] == 0)
                else:
                    indexes[name].columns.append(column)

        return indexes.values()

    def _get_table_columns(self, entity: Type[ffd.Entity]):
        result = self._execute(*self._generate_query(entity, f'{self._sql_prefix}/get_columns.sql'))
        ret = []
        if result:
            for row in result:
                ret.append(Column(name=row['COLUMN_NAME'], type=row['COLUMN_TYPE']))
        return ret



from __future__ import annotations

import firefly_aws.domain as domain


class DdbResourceMonitor(domain.ResourceMonitor):
    _ddb_deserializer: domain.DdbDeserializer = None
    _ddb_client = None
    _ddb_table: str = None

    def record_execution(self, message: str, seconds: float, memory: float) -> dict:
        response = self._ddb_client.update_item(
            TableName=self._ddb_table,
            Key={
                'pk': {'S': message},
                'sk': {'S': 'resource-usage'}
            },
            UpdateExpression=f"""
                SET memory = list_append(if_not_exists(memory, :empty), :memory),
                    times = list_append(if_not_exists(times, :empty), :times)
            """,
            ExpressionAttributeValues={
                ':memory': {'L': [{'N': str(memory)}]},
                ':times': {'L': [{'N': str(seconds)}]},
                ':empty': {'L': []}
            },
            ReturnValues='ALL_NEW'
        )

        return self._ddb_deserializer.deserialize(response['Attributes'])

    def get_execution_metrics(self, message: str):
        response = self._ddb_client.get_item(
            TableName=self._ddb_table,
            Key={
                'pk': {'S': message},
                'sk': {'S': 'resource-usage'}
            }
        )

        return self._ddb_deserializer.deserialize(response['Item']) if 'Item' in response else None

    def set_memory_level(self, message: str, memory: int):
        self._ddb_client.update_item(
            TableName=self._ddb_table,
            Key={
                'pk': {'S': message},
                'sk': {'S': 'memory-level'}
            },
            UpdateExpression=f"""
                SET memory_level = :memory
            """,
            ExpressionAttributeValues={
                ':memory': {'N': str(memory)},
            },
        )

    def get_memory_level(self, message: str):
        response = self._ddb_client.get_item(
            TableName=self._ddb_table,
            Key={
                'pk': {'S': message},
                'sk': {'S': 'memory-level'}
            }
        )

        return self._ddb_deserializer.deserialize(response['Item'])['memory_level'] if 'Item' in response else None



from __future__ import annotations

import bz2

import firefly as ff


class LoadPayload(ff.DomainService):
    _s3_client = None
    _serializer: ff.Serializer = None
    _bucket: str = None

    def __call__(self, key: str):
        response = self._s3_client.get_object(
            Bucket=self._bucket,
            Key=key
        )
        data = response['Body'].read()
        if key.endswith('bz2'):
            data = bz2.decompress(data)
        return self._serializer.deserialize(data)



from __future__ import annotations

import json
import os
import resource
from typing import Callable

import firefly as ff
import psutil
from botocore.exceptions import EndpointConnectionError
from firefly import domain as ffd

import firefly_aws.domain as domain

TIME_LIMIT = 900_000


if os.environ.get('ADAPTIVE_MEMORY'):
    # @ff.register_middleware(index=1, buses=['event', 'command'])
    class ResourceMonitoringMiddleware(ff.Middleware, domain.ResourceNameAware, ff.LoggerAware):
        _resource_monitor: domain.ResourceMonitor = None
        _execution_context: domain.ExecutionContext = None
        _message_factory: ff.MessageFactory = None
        _message_transport: ff.MessageTransport = None
        _configuration: ff.Configuration = None
        _requeue_message: domain.RequeueMessage = None
        _kinesis_client = None
        _context: str = None
        _memory_settings: list = None

        def __init__(self):
            context = self._configuration.contexts['firefly_aws']
            if context.get('memory_async') == 'adaptive':
                self._memory_settings = sorted(list(map(int, context.get('memory_settings'))))
                if self._memory_settings is None:
                    raise ff.ConfigurationError(
                        'When using "adaptive" memory you must provide a list of memory_settings'
                    )
                # self._set_soft_memory_limit()

        def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
            response = None

            try:
                response = next_(message)

                if isinstance(message, ff.Event) or (isinstance(message, ff.Command) and hasattr(message, '_async')):
                    memory_limit = self._execution_context.context.memory_limit_in_mb
                    memory_index = self._memory_settings.index(int(memory_limit))
                    memory_percent_used = psutil.Process(os.getpid()).memory_percent() / 100
                    memory_used = memory_percent_used * float(memory_limit)
                    self._kinesis_client.put_record(
                        StreamName=self._stream_resource_name(self._context),
                        Data=json.dumps({
                            'event_type': 'resource-usage',
                            'message': str(message),
                            'memory_used': float(memory_used),
                            'run_time': float(
                                TIME_LIMIT - self._execution_context.context.get_remaining_time_in_millis()
                            ),
                            'max_memory': float(memory_limit),
                            'prev_memory_tier': float(self._memory_settings[memory_index - 1])
                            if memory_index > 0 else None,
                        }).encode('utf-8'),
                        PartitionKey='resource-monitor'
                    )
            except (MemoryError, EndpointConnectionError, SystemExit):
                self._requeue_message(message)
            except AttributeError:
                pass

            return response

        def _set_soft_memory_limit(self):
            if not self._execution_context.context:
                return

            limit = int(self._execution_context.context.memory_limit_in_mb * 1048576 * .9)
            self.info(f'Setting soft memory limit to {limit} bytes')
            _, hard = resource.getrlimit(resource.RLIMIT_AS)
            resource.setrlimit(resource.RLIMIT_AS, (limit, hard))



import os


if os.environ.get('ADAPTIVE_MEMORY'):
    from .adaptive_memory_routing_middleware import AdaptiveMemoryRoutingMiddleware
    from .resource_monitoring_middleware import ResourceMonitoringMiddleware



from __future__ import annotations

from math import floor
from statistics import median
from typing import List

import firefly as ff


class FindOutlierThreshold(ff.DomainService):
    
    def __call__(self, memory_list: List[int]):
        sorted_memory_usage = sorted(memory_list)
        median_memory_usage = median(sorted_memory_usage)
        absolute_deviation_from_median = [abs(value - median_memory_usage) for value in sorted_memory_usage]
        median_absolute_deviation = median(absolute_deviation_from_median)
        outlier_threshold = 0
        
        # If more than 50% of our values are the same value, MAD will be 0. Then initiate backup plan with z_score
        # (very rare edge-case) Not the best case to use z-score, but good backup
        if median_absolute_deviation == 0:
            # Outlier Threshold at 99.5% of runs if MAD is 0
            ninety_nine_five_threshold = floor(0.995 * len(sorted_memory_usage))
            outlier_threshold = sorted_memory_usage[ninety_nine_five_threshold]
        else:
            # 1.4826 is a constant linked to assumption of normally distributed data excluding outliers
            memory_mad = float(median_absolute_deviation) * 1.4826

            for value in sorted_memory_usage:
                memory_mad_relative_deviation = float(value - median_memory_usage) / memory_mad
                # Any relative deviation greater than 3.0 is an outlier
                if abs(memory_mad_relative_deviation) < 3.0:
                    # since we're iterating through a sorted list, the value can always be updated if true
                    outlier_threshold = value

        return outlier_threshold



#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
from dataclasses import fields
from typing import Type, get_type_hints, List, Dict

import firefly as ff
from firefly import domain as ffd

from .data_api_mysql_base import DataApiMysqlBase


# noinspection PyDataclass
class DataApiMysqlMappedStorageInterface(DataApiMysqlBase):
    _sql_prefix = 'mysql'
    _identifier_quote_char = '`'
    _map_all = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _get_average_row_size(self, entity: Type[ff.Entity]):
        schema, table = self._fqtn(entity).split('.')
        sql = f"""
            select avg_row_length
            from information_schema.tables
            where table_schema = '{schema}'
            and table_name = '{table}'
        """
        result = ff.retry(lambda: self._execute(sql))
        try:
            return result[0]['AVG_ROW_LENGTH'] / 1024
        except KeyError:
            return 1



from __future__ import annotations

from datetime import date, datetime
from typing import Tuple, List

import firefly as ff
from botocore.exceptions import ClientError


class S3FileSystem(ff.FileSystem, ff.LoggerAware):
    _s3_client = None
    _bucket: str = None

    def read(self, file_name: str) -> ff.File:
        bucket, file_name = self._parse_file_path(file_name)
        try:
            response = self._s3_client.get_object(Bucket=bucket, Key=file_name)
        except self._s3_client.exceptions.NoSuchKey:
            raise ff.NoSuchFile()

        content = response['Body'].read()
        try:
            content = content.decode('utf-8')
        except UnicodeDecodeError:
            pass

        return ff.File(
            name=file_name,
            content=content,
            content_type=response.get('ContentType', None)
        )

    def write(self, file: ff.File, path: str = None):
        path = '/'.join([(path or '').rstrip('/'), file.name])
        bucket, file_name = self._parse_file_path(path)
        params = {}
        if file.content_type is not None:
            params['ContentType'] = file.content_type
        self._s3_client.put_object(
            Bucket=bucket,
            Key=file_name,
            Body=file.content,
            **params
        )

    def list(self, path: str) -> List[Tuple[str, dict]]:
        bucket, file_name = self._parse_file_path(path)
        params = {'Bucket': bucket, 'Prefix': file_name}
        ret = []

        while True:
            response = self._s3_client.list_objects_v2(**params)
            if 'Contents' in response:
                for item in response['Contents']:
                    ret.append((f"{bucket}/{item['Key']}", {
                        'size': item['Size'],
                        'last_modified': item['LastModified'],
                    }))
            if response['IsTruncated'] and 'NextContinuationToken' in response:
                params['ContinuationToken'] = response['NextContinuationToken']
            else:
                break

        return ret

    def filter(self, path: str, fields: list, criteria: ff.BinaryOp):
        bucket, file_name = self._parse_file_path(path)

        sql = f"select {', '.join(fields)} from s3object s"
        if criteria is not None:
            where, params = criteria.to_sql(prefix='s')
            for k, v in params.items():
                if isinstance(v, str):
                    where = where.replace(f':{k}', f"'{v}'")
                elif isinstance(v, (datetime, date)):
                    format_ = "y-MM-dd''T''H:m:ss" if isinstance(v, datetime) else 'y-MM-dd'
                    where = where.replace(f':{k}', f"TO_TIMESTAMP('{v.isoformat()}', '{format_}')")
                else:
                    where = where.replace(f':{k}', str(v))
            sql += f' where {where}'

        compression = 'NONE'
        fn = file_name.lower()
        if fn.endswith('.bz2'):
            compression = 'BZIP2'
        elif fn.endswith('.gz'):
            compression = 'GZIP'

        input_serialization = {
            'CompressionType': compression
        }
        if '.parquet' in fn:
            input_serialization['Parquet'] = {}
        elif '.json' in fn:
            input_serialization['JSON'] = {
                'Type': 'DOCUMENT',
            }
        elif '.csv' in fn:
            input_serialization['CSV'] = {
                'FileHeaderInfo': 'Use',
            }

        try:
            response = self._s3_client.select_object_content(
                Bucket=bucket,
                Key=file_name,
                InputSerialization=input_serialization,
                OutputSerialization={'JSON': {}},
                Expression=sql,
                ExpressionType='SQL'
            )
        except ClientError as e:
            print(e)
            return ''

        ret = '['
        for event in response['Payload']:
            if 'Records' in event:
                ret += event['Records']['Payload'].decode('utf-8').replace("\n", ",").replace('\\r', '')

        return ret.rstrip(',') + ']'

    def _parse_file_path(self, path: str):
        parts = path.lstrip('/').split('/')
        bucket = parts.pop(0)
        file_name = '/'.join(parts)

        return bucket, file_name



from __future__ import annotations

from datetime import datetime, timedelta
from time import sleep

import firefly as ff
from botocore.exceptions import ClientError


class DdbMutex(ff.Mutex):
    _ddb_client = None
    _ddb_table: str = None

    def acquire(self, key: str, timeout: int = None) -> bool:
        mark = datetime.now()
        while True:
            try:
                self._ddb_client.put_item(
                    TableName=self._ddb_table,
                    Item={
                        'pk': {'S': key},
                        'sk': {'S': 'mutex'},
                        'TimeToLive': {'N': str(round((datetime.now() + timedelta(seconds=(60 * 15))).timestamp()))},
                    },
                    ConditionExpression=f'attribute_not_exists(pk)'
                )
                return True
            except ClientError as e:
                if 'ConditionalCheckFailedException' in str(e):
                    if timeout is not None and (datetime.now() - mark).seconds >= timeout:
                        raise TimeoutError(f'Could not acquire mutex for key: {key}')
                    sleep(1)
                else:
                    raise e

    def release(self, key: str):
        try:
            self._ddb_client.delete_item(
                TableName=self._ddb_table,
                Key={
                    'pk': {'S': key},
                    'sk': {'S': 'mutex'},
                }
            )
            return True
        except ClientError:
            return False



from __future__ import annotations

import firefly as ff

import firefly_aws.domain as domain

TIME_LIMIT = 900_000


class RequeueMessage(ff.DomainService):
    _resource_monitor: domain.ResourceMonitor = None
    _execution_context: domain.ExecutionContext = None
    _message_transport: ff.MessageTransport = None
    _configuration: ff.Configuration = None
    _context: str = None

    def __init__(self):
        context = self._configuration.contexts['firefly_aws']
        if context.get('memory_async') == 'adaptive':
            self._memory_settings = sorted(list(map(int, context.get('memory_settings'))))
            if self._memory_settings is None:
                raise ff.ConfigurationError(
                    'When using "adaptive" memory you must provide a list of memory_settings'
                )

    def __call__(self, message: ff.Message, **kwargs):
        if not self._execution_context.context:
            return

        memory_index = None
        memory_limit = self._execution_context.context.memory_limit_in_mb

        if self._memory_settings is not None:
            memory_index = self._memory_settings.index(int(memory_limit))

        if memory_index is None or memory_index >= (len(self._memory_settings) - 1):
            raise MemoryError()
        setattr(message, '_memory', self._memory_settings[memory_index + 1])

        if isinstance(message, ff.Event):
            self._message_transport.dispatch(message)
        elif isinstance(message, ff.Command):
            setattr(message, '_async', True)
            self._message_transport.invoke(message)



from __future__ import annotations

from abc import ABC, abstractmethod


class ResourceMonitor(ABC):
    @abstractmethod
    def record_execution(self, message: str, seconds: float, memory: float):
        pass

    @abstractmethod
    def get_execution_metrics(self, message: str):
        pass

    @abstractmethod
    def set_memory_level(self, message: str, memory: int):
        pass

    @abstractmethod
    def get_memory_level(self, message: str):
        pass



#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List

import firefly as ff

from .service import Service
from .stack import Stack


class Project(ff.AggregateRoot):
    id: str = ff.id_(is_uuid=False)
    services: List[Service] = ff.list_()

    def plan_deployment(self, deployment: ff.Deployment, config: ff.Configuration):
        project = config.all.get('project')
        aws = config.contexts.get('firefly_aws')
        try:
            api_gateway_resource = aws.get('api_gateways').get('default')
        except AttributeError:
            api_gateway_resource = None

        for s in deployment.services:
            id_ = f'{project}-{s.name}'
            service = self.get_service(id_) or Service(id=id_)
            function = service.get_lambda()

            for gateway in s.api_gateways:
                for endpoint in gateway.endpoints:
                    print(endpoint.route)

            for topic in s.network_topology.topics:
                print(f'{topic.name}:')
                for sub in topic.subscribers:
                    print(f'  {sub.name}')

    def get_service(self, id_: str):
        for s in self.services:
            if s.id == id_:
                return s



from __future__ import annotations

from typing import Type

import firefly as ff
from firefly.infrastructure.repository.rdb_repository import Column, Index

from ..data_api_storage_interface import DataApiStorageInterface


class DataApiPgStorageInterface(DataApiStorageInterface):
    _sql_prefix = 'pg'

    def _get_table_indexes(self, entity: Type[ff.Entity]):
        schema, table = self._fqtn(entity).split('.')
        result = self._execute(*self._generate_query(entity, f'{self._sql_prefix}/get_indexes.sql', {
            'schema': schema,
            'table': table
        }))
        indexes = {}
        if result:
            for row in result:
                name = row['indexname']
                if name.endswith('_pkey'):
                    continue
                column = ''
                indexes[name] = Index(name=name, columns=[column], unique=('UNIQUE' in row['indexdef']))

        return indexes.values()

    def _get_table_columns(self, entity: Type[ff.Entity]):
        schema, table = self._fqtn(entity).split('.')
        result = self._execute(*self._generate_query(entity, f'{self._sql_prefix}/get_columns.sql', {
            'schema': schema,
            'table': table
        }))
        ret = []
        if result:
            for row in result:
                ret.append(Column(name=row['column_name'], type=row['data_type']))
        return ret

    def _get_average_row_size(self, entity: Type[ff.Entity]):
        result = self._execute(f"select CEIL(AVG(LENGTH(document::text)))::int as c from {self._fqtn(entity)}")
        try:
            return result[0]['c'] / 1024
        except KeyError:
            return 1

    @staticmethod
    def _cast_json():
        return '::json'

    @staticmethod
    def _cast_uuid():
        return '::uuid'
